// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import {ERC20} from "@openzeppelin/contracts/token/ERC20/ERC20.sol";
import "./Reputation.sol";
import "./Bench.sol";
import "./LightweightBench.sol";
import "../GlobalForum.sol";
import "../interfaces/IAcceptAvailability.sol";
import "../interfaces/IOnValidate.sol";

contract DAO {
    Reputation rep;
    GlobalForum forum;
    Bench bench;
    LightweightBench lightweightBench;
    mapping(uint => address) public members;
    uint public memberCount;
    mapping(address => bool) public isMember;

    event PostAdded(string id);
    event ValidationPoolInitiated(uint poolIndex);
    event ValidationPoolResolved(
        uint poolIndex,
        bool votePasses,
        bool quorumMet
    );
    event LWValidationPoolInitiated(uint poolIndex);
    event LWValidationPoolResolved(
        uint poolIndex,
        bool votePasses,
        bool quorumMet
    );
    event LWResultProposed(
        uint poolIndex,
        uint proposedResultIndex,
        string proposedResultHash
    );

    constructor(
        Reputation reputation_,
        Bench bench_,
        LightweightBench lightweightBench_,
        GlobalForum forum_
    ) {
        rep = reputation_;
        bench = bench_;
        lightweightBench = lightweightBench_;
        forum = forum_;
        rep.registerDAO(this);
        bench.registerDAO(this, forum);
        lightweightBench.registerDAO(this);
    }

    function emitPostAdded(string calldata id) public {
        emit PostAdded(id);
    }

    function emitValidationPoolInitiated(uint poolIndex) public {
        emit ValidationPoolInitiated(poolIndex);
    }

    function emitValidationPoolResolved(
        uint poolIndex,
        bool votePasses,
        bool quorumMet
    ) public {
        emit ValidationPoolResolved(poolIndex, votePasses, quorumMet);
    }

    function emitLWValidationPoolInitiated(uint poolIndex) public {
        emit LWValidationPoolInitiated(poolIndex);
    }

    function emitLWResultProposed(
        uint poolIndex,
        uint proposedResultIndex,
        string calldata proposedResultHash
    ) public {
        emit LWResultProposed(
            poolIndex,
            proposedResultIndex,
            proposedResultHash
        );
    }

    function update(address from, address to, uint256 value) public {
        require(
            msg.sender == address(lightweightBench) ||
                msg.sender == address(bench),
            "Only DAO core contracts may call update"
        );
        rep.update(from, to, value);
    }

    function mint(address account, uint256 value) public {
        require(
            msg.sender == address(lightweightBench) ||
                msg.sender == address(bench),
            "Only DAO core contracts may call mint"
        );
        rep.mint(account, value);
    }

    function burn(address account, uint256 value) public {
        require(
            msg.sender == address(lightweightBench) ||
                msg.sender == address(bench),
            "Only DAO core contracts may call burn"
        );
        rep.burn(account, value);
    }

    function registerMember(address account) public {
        require(
            msg.sender == address(lightweightBench) ||
                msg.sender == address(bench),
            "Only DAO core contracts may call registerMember"
        );
        if (!isMember[account]) {
            members[memberCount++] = account;
            isMember[account] = true;
        }
    }

    function balanceOf(address account) public view returns (uint256) {
        return rep.balanceOf(account);
    }

    function totalSupply() public view returns (uint256) {
        return rep.totalSupply();
    }

    function allowance(
        address owner,
        address spender
    ) public view returns (uint256) {
        return rep.allowance(owner, spender);
    }

    function forwardAllowance(
        address owner,
        address to,
        uint256 amount
    ) public {
        rep.spendAllowance(owner, msg.sender, amount);
        rep.approve(owner, to, rep.allowance(owner, to) + amount);
    }

    /// Authorize a contract to transfer REP, and call that contract's acceptAvailability method
    function stakeAvailability(
        address to,
        uint256 value,
        uint duration
    ) external returns (bool) {
        uint refund = IAcceptAvailability(to).acceptAvailability(
            msg.sender,
            value,
            duration
        );
        rep.approve(
            msg.sender,
            to,
            rep.allowance(msg.sender, to) + value - refund
        );
        return true;
    }

    function distributeFeeAmongMembers() public payable {
        uint allocated;
        for (uint i = 0; i < memberCount; i++) {
            address member = members[i];
            uint share;
            if (i < memberCount - 1) {
                share = (msg.value * balanceOf(member)) / totalSupply();
                allocated += share;
            } else {
                // Due to rounding, give the remainder to the last member
                share = msg.value - allocated;
            }
            // TODO: For efficiency this could be modified to hold the funds for recipients to withdraw
            payable(member).transfer(share);
        }
    }

    function getValidationPool(
        uint poolIndex
    )
        public
        view
        returns (
            uint id,
            address sender,
            uint stakeCount,
            ValidationPoolParams memory params,
            ValidationPoolProps memory props,
            bool callbackOnValidate,
            bytes memory callbackData
        )
    {
        return bench.validationPools(poolIndex);
    }

    function getValidationPoolCount() public view returns (uint) {
        return bench.validationPoolCount();
    }

    function initiateValidationPool(
        string calldata postId,
        uint duration,
        uint[2] calldata quorum, // [Numerator, Denominator]
        uint[2] calldata winRatio, // [Numerator, Denominator]
        uint bindingPercent,
        bool redistributeLosingStakes,
        bool callbackOnValidate,
        bytes calldata callbackData
    ) external payable returns (uint) {
        return
            bench.initiateValidationPool{value: msg.value}(
                msg.sender,
                postId,
                duration,
                quorum,
                winRatio,
                bindingPercent,
                redistributeLosingStakes,
                callbackOnValidate,
                callbackData
            );
    }

    function stakeOnValidationPool(
        uint poolIndex,
        uint256 amount,
        bool inFavor
    ) public {
        require(
            balanceOf(msg.sender) >= amount,
            "Insufficient REP balance to cover stake"
        );
        bench.stakeOnValidationPool(poolIndex, msg.sender, amount, inFavor);
    }

    /// Accept reputation stakes toward a validation pool
    function delegatedStakeOnValidationPool(
        uint poolIndex,
        address owner,
        uint256 amount,
        bool inFavor
    ) public {
        if (allowance(owner, msg.sender) < amount) {
            amount = allowance(owner, msg.sender);
        }
        rep.spendAllowance(owner, msg.sender, amount);
        bench.stakeOnValidationPool(poolIndex, owner, amount, inFavor);
    }

    function evaluateOutcome(uint poolIndex) public returns (bool) {
        return bench.evaluateOutcome(poolIndex);
    }

    function getLWValidationPool(
        uint poolIndex
    )
        public
        view
        returns (
            uint id,
            address sender,
            uint stakeCount,
            LWVPoolParams memory params,
            LWVPoolProps memory props,
            bool callbackOnValidate,
            bytes memory callbackData
        )
    {
        return lightweightBench.validationPools(poolIndex);
    }

    function getLWValidationPoolCount() public view returns (uint) {
        return lightweightBench.validationPoolCount();
    }

    function initiateLWValidationPool(
        string calldata postId,
        uint duration,
        uint[2] calldata quorum, // [Numerator, Denominator]
        uint[2] calldata winRatio, // [Numerator, Denominator]
        uint bindingPercent,
        bool redistributeLosingStakes,
        bool callbackOnValidate,
        bytes calldata callbackData
    ) external payable returns (uint) {
        return
            lightweightBench.initiateValidationPool{value: msg.value}(
                msg.sender,
                postId,
                duration,
                quorum,
                winRatio,
                bindingPercent,
                redistributeLosingStakes,
                callbackOnValidate,
                callbackData
            );
    }

    function proposeLWResult(
        uint poolIndex,
        string calldata resultHash,
        Transfer[] calldata transfers
    ) external {
        lightweightBench.proposeResult(poolIndex, resultHash, transfers);
    }

    function stakeOnLWValidationPool(
        uint poolIndex,
        string calldata resultHash,
        uint256 amount,
        bool inFavor
    ) public {
        require(
            balanceOf(msg.sender) >= amount,
            "Insufficient REP balance to cover stake"
        );
        lightweightBench.stakeOnValidationPool(
            poolIndex,
            resultHash,
            msg.sender,
            amount,
            inFavor
        );
    }

    /// Accept reputation stakes toward a validation pool
    function delegatedStakeOnLWValidationPool(
        uint poolIndex,
        string calldata resultHash,
        address owner,
        uint256 amount,
        bool inFavor
    ) public {
        if (allowance(owner, msg.sender) < amount) {
            amount = allowance(owner, msg.sender);
        }
        rep.spendAllowance(owner, msg.sender, amount);
        lightweightBench.stakeOnValidationPool(
            poolIndex,
            resultHash,
            owner,
            amount,
            inFavor
        );
    }

    function evaluateLWOutcome(uint poolIndex) public returns (bool) {
        return lightweightBench.evaluateOutcome(poolIndex);
    }

    function onValidate(
        address target,
        bool votePasses,
        bool quorumMet,
        uint stakedFor,
        uint stakedAgainst,
        bytes calldata callbackData
    ) public {
        require(
            msg.sender == address(lightweightBench) ||
                msg.sender == address(bench),
            "Only DAO core contracts may call onValidate"
        );
        IOnValidate(target).onValidate(
            votePasses,
            quorumMet,
            stakedFor,
            stakedAgainst,
            callbackData
        );
    }
}

/// Convenience contract to extend for other contracts that will be initialized to
/// interact with a DAO contract.
contract DAOContract {
    DAO immutable dao;

    constructor(DAO dao_) {
        dao = dao_;
    }
}
